#ifndef TERMINOLOGY_ABOUT_H_
#define TERMINOLOGY_ABOUT_H_ 1

void about_show(Evas_Object *win, Evas_Object *base, Evas_Object *term,
                  void (*donecb) (void *data), void *donedata);

#endif
