#ifndef TERMINOLOGY_TERMPTY_EXT_H_
#define TERMINOLOGY_TERMPTY_EXT_H_ 1

Eina_Bool
termpty_ext_handle(Termpty *ty,
                   const Eina_Unicode *buf,
                   size_t blen);

#endif
